
//----------------------------------------------------------------------------
// StarChase -- an interstellar action extravaganza 
// by Christopher Rasmussen
// v 1.0 (c) 2010
//----------------------------------------------------------------------------

#include <iostream>
#include <cstdlib>
#include <time.h>
#include <ncurses.h>

using namespace std;

//----------------------------------------------------------------------------

const int SCREEN_WIDTH     = 20;
const int SCREEN_HEIGHT    = 15;
const int NUM_STARS        = 20;
const int OSCILLATOR_INDEX = 0;

const char FORWARD_KEY     = 'w';
const char REVERSE_KEY     = 's';
const char LEFT_KEY        = 'a';
const char RIGHT_KEY       = 'd';

// color 0 is reserved

const int SHIP_COLOR       = 1;
const int STAR_COLOR       = 2;
const int OSCILLATOR_COLOR = 3;

// global variable alert!!

bool using_color = false;

// this stuff probably should go in a header

void init_random();
int rand_low_high(int, int);
void init_ncurses();

void set_random_star_positions(int[], int[], int);
void draw_stars(int[], int[], bool[], int);
void draw_boundaries();
void draw_ship(int, int, int, int);
bool in_boundaries(int, int);
int check_for_star_collection(int, int, int[], int[], bool[], int);
void draw_stats(int, int);

//----------------------------------------------------------------------------

// initialize random number generation so that we get different
// sequences each time we run the program

void init_rand()
{
  time_t now;

  now = time(NULL);
  srand(now);
}

//----------------------------------------------------------------------------

// generate random number between low and high, inclusive

int rand_low_high(int low, int high)
{
  return low + rand() % (high + 1 - low);
}

//----------------------------------------------------------------------------

// get ready to draw with ncurses library

void init_ncurses()
{
  initscr();

  // comment this out to default to no colors
  
  if (has_colors()) {
    using_color = true;
    start_color();
    init_pair(SHIP_COLOR, COLOR_RED, COLOR_BLACK);
  }

  raw();
  noecho();
  curs_set(0);   // make cursor invisible
}

//----------------------------------------------------------------------------

// distribute stars randomly and say none have been collected yet
// some may be on top of each other

void initialize_stars(int star_x[], int star_y[], bool star_collected[], int num_stars)
{
  int i;

  for (i = 0; i < num_stars; i++) {
    star_x[i] = rand_low_high(1, SCREEN_WIDTH - 2);
    star_y[i] = rand_low_high(1, SCREEN_HEIGHT - 2);
    star_collected[i] = false;
  }
}

//----------------------------------------------------------------------------

// outline edges of playing area

void draw_boundaries()
{
  int i;

  // top and bottom

  for (i = 0; i < SCREEN_WIDTH; i++) {
    move(0, i);
    printw(".");
    move(SCREEN_HEIGHT - 1, i);
    printw(".");
  }

  // left and right

  for (i = 0; i < SCREEN_HEIGHT; i++) {
    move(i, 0);
    printw(".");
    move(i, SCREEN_WIDTH - 1);
    printw(".");
  }
 
}

//----------------------------------------------------------------------------

// show all remaining stars

void draw_stars(int x[], int y[], bool collected[], int n)
{
  int i;

  for (i = 0; i < n; i++) 
    if (!collected[i]) {
      move(y[i], x[i]);
      if (i != OSCILLATOR_INDEX)
	printw("*");
      else
	printw("X");
    }
}

//----------------------------------------------------------------------------

// show ship with current location and heading

void draw_ship(int x, int y, int dx, int dy)
{
  if (using_color) 
    attron(COLOR_PAIR(SHIP_COLOR));  

  move(y, x);
  printw("0");
  move(y + dy, x + dx);
  printw("+");

  if (using_color) 
    attroff(COLOR_PAIR(SHIP_COLOR));
}

//----------------------------------------------------------------------------

// based on key input, move ship as commanded

void control_ship_from_keyboard(char c, int & x, int & y, int & dx, int & dy)
{
  int temp;

  switch (c) {
    
    // move forward				
      
  case FORWARD_KEY:
    if (in_boundaries(x + dx, y + dy)) {
      x += dx;
      y += dy;
    }
    break;
    
    // move backward
    
  case REVERSE_KEY:
    if (in_boundaries(x - dx, y - dy)) {
      x -= dx;
      y -= dy;
    }
    break;

    // rotate left 90 degs.

  case LEFT_KEY:
    temp = -dx;
    dx = dy;
    dy = temp;
    break;

    // rotate right 90 degs.

  case RIGHT_KEY:
    temp = dx;
    dx = -dy;
    dy = temp;
    break;
  }

}

//----------------------------------------------------------------------------

// check that we are not touching the "edge of space"

bool in_boundaries(int x, int y)
{
  return (x >= 1 && x < (SCREEN_WIDTH - 1) &&
	  y >= 1 && y < (SCREEN_HEIGHT - 1));
}

//----------------------------------------------------------------------------

// loop over all uncollected stars and see which ones we're on top of.
// change collection status of any such stars and return how many we got.

int check_for_star_collection(int x, int y, int star_x[], int star_y[], bool star_collected[], int num_stars)
{
  int i, num_collections;

  for (i = 0, num_collections = 0; i < num_stars; i++) 
    if (!star_collected[i] && x == star_x[i] && y == star_y[i]) {
      star_collected[i] = true;
      num_collections++;
    }

  return num_collections;
}

//----------------------------------------------------------------------------

// draw user's score and time elapsed

void draw_stats(int num_stars_collected, int time_elapsed)
{
  move(SCREEN_HEIGHT, 0);
  printw("%3i s  %2i/%i stars", time_elapsed, num_stars_collected, NUM_STARS);

}

//----------------------------------------------------------------------------

// print title, instructions

void draw_splash()
{
  printw("Welcome to StarChase, the wacky game of energy collection!!\n");
  printw("Try to visit all the stars as fast as possible...\n");
  printw("The X is an oscillating star...it's not dangerous--just a little harder to get\n");
  printw("%c/%c will move your ship forward/backward; %c/%c will turn your ship left/right\n", FORWARD_KEY, REVERSE_KEY, LEFT_KEY, RIGHT_KEY);   
  printw("Hit 'e' to exit at any time\n");   
  printw("Ready?\n");   
}

//----------------------------------------------------------------------------

int main(void)
{
  int c;                                  // hold user's latest keyboard input

  int ship_x, ship_y, ship_dx, ship_dy;   // spaceship's current position and heading
  int star_position_x[NUM_STARS];         // locations of all stars
  int star_position_y[NUM_STARS];
  bool star_collected[NUM_STARS];         // whether star has been visited yet
  int num_stars_collected;                // how many stars user has so far

  init_rand();
  init_ncurses();

  draw_splash();
  refresh();			                                              
  c = getch();			                                              
  
  // set initial ship position and direction, star positions, and stats to zero

  initialize_stars(star_position_x, star_position_y, star_collected, NUM_STARS);

  ship_x = SCREEN_WIDTH/2;
  ship_y = SCREEN_HEIGHT/2;
  ship_dx = 0;
  ship_dy = -1;

  num_stars_collected = 0;

  time_t start_time = time(NULL);
  time_t now_time;
  int time_elapsed;

  timeout(100);

  // loop to allow ship motion

  while (c != 'e') {

    now_time = time(NULL);
    time_elapsed = static_cast<int>(now_time - start_time);

    // one star oscillates

    int rand_dx = rand_low_high(-1, 1);
    int rand_dy = rand_low_high(-1, 1);

    if (in_boundaries(star_position_x[OSCILLATOR_INDEX] + rand_dx, star_position_y[OSCILLATOR_INDEX] + rand_dy)) {
      star_position_x[OSCILLATOR_INDEX] += rand_dx;
      star_position_y[OSCILLATOR_INDEX] += rand_dy;
    }

    // clear screen & print it in current location

    erase();

    draw_boundaries();
    draw_stars(star_position_x, star_position_y, star_collected, NUM_STARS);
    draw_ship(ship_x, ship_y, ship_dx, ship_dy);
    draw_stats(num_stars_collected, time_elapsed);

    refresh();

    // move ship

    c = getch();
    control_ship_from_keyboard(c, ship_x, ship_y, ship_dx, ship_dy);

    // check for star collection and update stats

    num_stars_collected += check_for_star_collection(ship_x, ship_y, star_position_x, star_position_y, star_collected, NUM_STARS);

    if (num_stars_collected == NUM_STARS)
      break;
  }

  // get out of ncurses mode so terminal is not screwed up when program exits

  endwin();

  // victory message?

  if (num_stars_collected == NUM_STARS)
    cout << "You got all of the stars in " << time_elapsed << " seconds\n";
  else
    cout << "Thanks for playing...\n";

  exit(1);	
}

//----------------------------------------------------------------------------
//----------------------------------------------------------------------------
